@@ -2,7 +2,12 @@ use heck::ToSnakeCase;
 use proc_macro::TokenStream;
 use proc_macro_error::abort;
 use quote::{format_ident, quote, ToTokens};
-use syn::{FnArg, Ident, ImplItem, ItemImpl, ItemStruct, LitStr, PatType, ReturnType, Type};
+use syn::{FnArg, Ident, ImplItem, ItemImpl, ItemStruct, PatType, ReturnType, Type, Block};
+
+#[proc_macro]
+pub fn top_route(_: TokenStream) -> TokenStream {
+    quote!(__srpc_inner_ctx).into()
+}
 
 #[proc_macro_attribute]
 #[proc_macro_error::proc_macro_error]
@@ -30,14 +35,8 @@ pub fn rpc(attrs: TokenStream, item: TokenStream) -> TokenStream {
     }
 }
 
-fn struct_rpc(attrs: TokenStream, item: ItemStruct) -> TokenStream {
+fn struct_rpc(_: TokenStream, item: ItemStruct) -> TokenStream {
     let ident = &item.ident;
-    let endpoint = syn::parse::<LitStr>(attrs)
-        .and_then(|s| Ok(LitStr::new(&s.value().to_snake_case(), s.span())))
-        .unwrap_or(LitStr::new(
-            &ident.clone().to_string().to_snake_case(),
-            ident.span(),
-        ));
     let vis = &item.vis;
     let peer_name = format_ident!("{}Peer", &ident);
     let (implgen, tygen, whre) = item.generics.split_for_impl();
@@ -46,16 +45,17 @@ fn struct_rpc(attrs: TokenStream, item: ItemStruct) -> TokenStream {
 
     quote!(
         #item
-        #[cfg(not(target_arch = "wasm32"))]
-        impl #implgen ::srpc::canary::routes::RegisterEndpoint for #ident #tygen #whre {
-            const ENDPOINT: &'static str = #endpoint;
-        }
-        #vis struct #peer_name #tygen #whre (pub ::srpc::canary::Channel, ::core::marker::PhantomData<( #(#ty_params),* )>);
-        impl #implgen From<::srpc::canary::Channel> for #peer_name #tygen #whre {
-            fn from(c: ::srpc::canary::Channel) -> Self {
+        #vis struct #peer_name #tygen #whre (pub ::srpc::__private::canary::Channel, ::core::marker::PhantomData<( #(#ty_params),* )>);
+        impl #implgen From<::srpc::__private::canary::Channel> for #peer_name #tygen #whre {
+            fn from(c: ::srpc::__private::canary::Channel) -> Self {
                 #peer_name(c, ::core::marker::PhantomData::default())
             }
         }
+        impl #implgen From<#peer_name #tygen> for ::srpc::__private::canary::Channel #whre {
+            fn from(c: #peer_name #tygen) -> Self {
+                c.0
+            }
+        }
         impl #implgen ::srpc::Peer for #ident #tygen #whre {
             type Struct = #peer_name #tygen;
         }
@@ -63,18 +63,23 @@ fn struct_rpc(attrs: TokenStream, item: ItemStruct) -> TokenStream {
     .into()
 }
 
+#[derive(Clone)]
 struct Method<'a> {
     ident: &'a Ident,
     output: Option<&'a Box<Type>>,
     inputs: Option<Vec<&'a PatType>>,
     mutable: bool,
     consume: MethodKind,
+    block: &'a Block
 }
 
+#[derive(Clone)]
 enum MethodKind {
     Normal,  // normal rpc
     Consume, // consume channel
     Manual,  // reintroduce
+    Server,  // similar to consume, but does not create a client method
+    Client,  // create a client method
 }
 
 enum RpcProvider {
@@ -122,10 +127,10 @@ impl RpcProvider {
     fn get_type<T: ToTokens>(&self, top: T) -> quote::__private::TokenStream {
         match self {
             RpcProvider::RwLock => quote! {
-                ::srpc::RwLock<#top>
+                ::srpc::__private::RwLock<#top>
             },
             RpcProvider::Mutex => quote! {
-                ::srpc::Mutex<#top>
+                ::srpc::__private::Mutex<#top>
             },
             RpcProvider::Ref => quote! {
                 #top
@@ -168,6 +173,7 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
     let top_type_meta = provider.get_type(&top_type);
     let mut method_names = methods.iter().map(|s| &s.sig.ident).collect::<Vec<_>>();
     method_names.sort();
+    method_names.dedup();
     let repr = {
         match method_names.len() {
             0 => quote! { #[derive(::srpc::__private::Serialize, ::srpc::__private::Deserialize)] },
@@ -203,6 +209,7 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
             ReturnType::Default => None,
             ReturnType::Type(_, ty) => Some(ty),
         };
+        let block = &s.block;
 
         let mut mutable = false;
         let consume = s
@@ -213,14 +220,25 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
             .attrs
             .iter()
             .any(|attr| quote!(#[manual]).to_string() == quote!(#attr).to_string());
-        let consume = match (consume, manual) {
-            (false, true) => MethodKind::Manual,
-            (true, false) => MethodKind::Consume,
-            (false, false) => MethodKind::Normal,
-            (true, true) => abort!(
+        let server = s
+            .attrs
+            .iter()
+            .any(|attr| quote!(#[server]).to_string() == quote!(#attr).to_string());
+        let client = s
+            .attrs
+            .iter()
+            .any(|attr| quote!(#[client]).to_string() == quote!(#attr).to_string());
+        let consume = match (consume, manual, client, server) {
+            (false, true, false, false) => MethodKind::Manual,
+            (false, false, true, false) => MethodKind::Client,
+            (false, false, false, true) => MethodKind::Server,
+            (true, false, false, false) => MethodKind::Consume,
+            (false, false, false, false) => MethodKind::Normal,
+            (consume, manual, client, server) if consume as u8 + manual as u8 + client as u8 + server as u8 > 1 => abort!(
                 &s.sig.ident.span(),
-                "cannot have a method with consume and manual"
+                "cannot have a method with more than one modifier"
             ),
+            (_, _, _, _) => unreachable!(),
         };
 
         let iter = s
@@ -242,11 +260,13 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
             inputs,
             mutable,
             consume,
+            block
         }
     });
 
     let meta_ident = format_ident!("__srpc_inner_meta");
     let channel_ident = format_ident!("__srpc_inner_channel");
+    let context_ident = format_ident!("__srpc_inner_ctx");
 
     let matches = methods.clone().map(|method| {
         let ident = method.ident;
@@ -257,12 +277,12 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
         match (inputs, method.output, method.consume) {
             (None, None, MethodKind::Normal) => quote!(
                 __srpc_action::#ident => {
-                    #meta.#ident().await;
+                    #meta.#ident(&#context_ident).await;
                 }
             ),
             (None, Some(_), MethodKind::Normal) => quote!(
                 __srpc_action::#ident => {
-                    let res = #meta.#ident().await;
+                    let res = #meta.#ident(&#context_ident).await;
                     #channel_ident.send(res).await?;
                 }
             ),
@@ -280,7 +300,7 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
                     __srpc_action::#ident => {
                         #[allow(unused_parens)]
                         let #inputs = #channel_ident.receive().await?;
-                        #meta.#ident(#(#args),*).await;
+                        #meta.#ident(#(#args),* , &#context_ident).await;
                     }
                 )
             },
@@ -296,7 +316,7 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
                     __srpc_action::#ident => {
                         #[allow(unused_parens)]
                         let #inputs = #channel_ident.receive().await?;
-                        let res = #meta.#ident(#(#args),*).await;
+                        let res = #meta.#ident(#(#args),* , &#context_ident).await;
                         #channel_ident.send(res).await?;
                     }
                 )
@@ -307,7 +327,7 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
                 }
                 quote! {
                     __srpc_action::#ident => {
-                        return #meta.#ident(#channel_ident).await;
+                        return #meta.#ident(#channel_ident, &#context_ident).await;
                     }
                 }
             }
@@ -317,13 +337,33 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
                 }
                 quote! {
                     __srpc_action::#ident => {
-                        match #meta.#ident(#channel_ident).await {
+                        match #meta.#ident(#channel_ident, &#context_ident).await {
                             Ok(chan) => #channel_ident = chan,
                             Err(e) => return Err(e),
                         }
                     }
                 }
             },
+            (Some(inputs), Some(_), MethodKind::Server) => {
+                if inputs.len() != 1 {
+                    abort!(method.ident.span(), "manual methods can only have one argument with type Channel and return a canary::Result<Channel>")
+                }
+                quote! {
+                    __srpc_action::#ident => {
+                        match #meta.#ident(&mut #channel_ident, &#context_ident).await {
+                            Ok(_) => (),
+                            Err(e) => return Err(e),
+                        }
+                    }
+                }
+            },
+            (None, None, MethodKind::Client) => quote!(),
+            (None, Some(_), MethodKind::Client) => quote!(),
+            (Some(_), None, MethodKind::Client) => quote!(),
+            (Some(_), Some(_), MethodKind::Client) => quote!(),
+            (Some(_), None, MethodKind::Server) => abort!(method.ident.span(), "server methods need an argument with type Channel and return a canary::Result<Channel>"),
+            (None, Some(_), MethodKind::Server) => abort!(method.ident.span(), "server methods need an argument with type Channel and return a canary::Result<Channel>"),
+            (None, None, MethodKind::Server) => abort!(method.ident.span(), "server methods need an argument with type Channel and return a canary::Result<Channel>"),
             (None, None, MethodKind::Consume) => abort!(method.ident.span(), "methods that consume need an argument with type Channel and return a canary::Result<()>"),
             (None, Some(_), MethodKind::Consume) => abort!(method.ident.span(), "methods that consume can only return a canary::Result<()>"),
             (Some(_), None, MethodKind::Consume) => abort!(method.ident.span(), "methods that consume can only return a canary::Result<()>"),
@@ -352,36 +392,16 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
     let function = {
         let impl_generics = impl_generics.clone();
         quote! {
-            impl #impl_generics ::srpc::canary::service::Service for #top_type #whr {
+            impl #impl_generics ::srpc::__private::canary::service::Service for #top_type #whr {
                 const ENDPOINT: &'static str = #endpoint;
                 type Pipeline = ();
                 type Meta = ::std::sync::Arc<#top_type_meta>;
                 fn service(
                     #meta_ident: ::std::sync::Arc<#top_type_meta>,
-                ) -> Box<dyn Fn(::srpc::canary::igcp::BareChannel) + Send + Sync + 'static> {
-                    ::canary::service::run_metadata(#meta_ident, |#meta_ident: ::std::sync::Arc<#top_type_meta>, mut #channel_ident: ::srpc::canary::Channel| async move {
-                        loop {
-                            match #channel_ident.receive::<__srpc_action>().await? {
-                                #(#matches),*
-                            }
-                        }
-                    })
-                }
-            }
-        }
-    };
-
-    let static_function = {
-        let impl_generics = impl_generics.clone();
-        quote! {
-            impl #impl_generics ::srpc::canary::service::StaticService for #top_type #whr {
-                type Meta = ::std::sync::Arc<#top_type_meta>;
-                type Chan = ::srpc::canary::Channel;
-                fn introduce(
-                    #meta_ident: ::std::sync::Arc<#top_type_meta>,
-                    mut #channel_ident: ::srpc::canary::Channel,
-                ) -> ::srpc::canary::runtime::JoinHandle<::srpc::canary::Result<()>> {
-                    ::srpc::canary::runtime::spawn(async move {
+                ) -> ::srpc::__private::canary::service::Svc {
+                    ::srpc::__private::canary::service::run_metadata(
+                        #meta_ident,
+                        |#meta_ident: ::std::sync::Arc<#top_type_meta>, mut #channel_ident: ::srpc::__private::canary::Channel, #context_ident: ::srpc::__private::canary::Ctx| async move {
                         loop {
                             match #channel_ident.receive::<__srpc_action>().await? {
                                 #(#matches),*
@@ -399,16 +419,16 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
         let has_output = method.output.is_some();
         let has_input = method.inputs.is_some();
 
-        let result = method.output
+        let res = method.output
             .map(|s| *s.clone())
             .unwrap_or(syn::parse2(quote!(())).unwrap());
-        let result = quote!( ::srpc::canary::Result<#result> );
+        let result = quote!( ::srpc::__private::canary::Result<#res> );
 
         let inputs = method.inputs.unwrap_or_default();
         let mut params = vec![];
         let name = method.ident;
 
-        let args = inputs.into_iter().map(|inp| {
+        let args = inputs.clone().into_iter().map(|inp| {
             let mut inp = inp.clone();
             let ty = inp.ty;
             let ty: Type = syn::parse2(quote!(impl std::borrow::Borrow<#ty>)).unwrap();
@@ -421,7 +441,7 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
             ret
         }).collect::<Vec<_>>();
 
-        match (has_output, has_input, method.consume) {
+        match (has_output, has_input, &method.consume) {
             (true, true, MethodKind::Normal) => quote! {
                 pub async fn #name(&mut self #(,#args)*) -> #result {
                     self.0.send(__srpc_action::#name).await?;
@@ -451,17 +471,37 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
                 }
             },
             (true, true, MethodKind::Consume) => quote! {
-                pub async fn #name(mut self) -> ::srpc::canary::Result<::srpc::canary::Channel> {
+                pub async fn #name(mut self) -> ::srpc::__private::canary::Result<::srpc::__private::canary::Channel> {
                     self.0.send(__srpc_action::#name).await?;
                     Ok(self.0)
                 }
             },
             (true, true, MethodKind::Manual) => quote! {
-                pub async fn #name(mut self) -> ::srpc::canary::Result<::srpc::canary::Channel> {
+                pub async fn #name(mut self) -> ::srpc::__private::canary::Result<::srpc::__private::canary::Channel> {
                     self.0.send(__srpc_action::#name).await?;
                     Ok(self.0)
                 }
             },
+            (_, true, MethodKind::Client) => {
+                let block = method.block;
+                let mut inputs = inputs.clone();
+                let chan_ident = &inputs.first().take().unwrap().pat;
+                inputs.remove(0);
+                quote! { // todo!
+                    pub async fn #name(&mut self #(,#inputs)*) -> #res {
+                        self.0.send(__srpc_action::#name).await?;
+
+                        let #chan_ident = &mut self.0;
+
+                        let __private_srpc_inner_result = {
+                            #block
+                        };
+                        __private_srpc_inner_result
+                    }
+                }
+            },
+            (_, _, MethodKind::Server) => quote!(),
+            (_, _, MethodKind::Client) => abort!(method.ident.span(), "client methods can only have an argument with type &mut Channel"),
             (true, false, MethodKind::Consume) => abort!(method.ident.span(), "methods that consume can only have an argument with type Channel and return a canary::Result<()>"),
             (false, true, MethodKind::Consume) => abort!(method.ident.span(), "methods that consume can only have an argument with type Channel and return a canary::Result<()>"),
             (false, false, MethodKind::Consume) => abort!(method.ident.span(), "methods that consume can only have an argument with type Channel and return a canary::Result<()>"),
@@ -477,23 +517,39 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
         }
     };
 
-    item.items
-        .iter_mut()
-        .map(|s| {
-            if let ImplItem::Method(method) = s {
+    let items = item.items
+        .into_iter()
+        .filter(|s| {
+            let mut stay = true;
+            if let ImplItem::Method(method) = &s {
+                let attrs = &method.attrs;
+                for attr in attrs {
+                    if quote!(#[client]).to_string() == quote!(#attr).to_string() {
+                        stay = false;
+                    }
+                }
+            }
+            stay
+        })
+        .map(|mut s| {
+            if let ImplItem::Method(method) = &mut s {
                 let attrs = method.attrs.clone();
                 let mut new_attrs = vec![];
                 for attr in attrs {
                     if quote!(#[consume]).to_string() != quote!(#attr).to_string()
                         && quote!(#[manual]).to_string() != quote!(#attr).to_string()
+                        && quote!(#[server]).to_string() != quote!(#attr).to_string()
+                        && quote!(#[client]).to_string() != quote!(#attr).to_string()
                     {
                         new_attrs.push(attr)
                     }
                 }
                 method.attrs = new_attrs;
             }
+            s
         })
-        .for_each(drop);
+        .collect::<Vec<_>>();
+    item.items = items;
     item.generics
         .type_params_mut()
         .map(|s| {
@@ -504,14 +560,20 @@ fn impl_rpc_provider(provider: RpcProvider, mut item: ItemImpl) -> TokenStream {
         })
         .for_each(drop);
 
+    item.items.iter_mut().map(|f| {
+        if let ImplItem::Method(f) = f {
+            let typed = syn::parse2(quote!(#context_ident: &::srpc::__private::canary::Ctx)).unwrap();
+            f.sig.inputs.push(typed);
+        }
+    }).for_each(drop);
+
     let ts = quote!(
         const _: () = {
+            #[cfg(not(target_arch = "wasm32"))]
             #item
             #enum_repr
             #[cfg(not(target_arch = "wasm32"))]
             #function
-            #[cfg(not(target_arch = "wasm32"))]
-            #static_function
             #peer_impl
         };
     );
